#!/bin/bash
#
# This script prepares keys and authenticated/signed secure boot
# setup information. The output will be enrolled in efi variables
# if run with the -e option.
#
# After a successful run, the output directory should contain:
# db_entry.esl
# KEK.esl
# PK.auth
# PK.esl
# PK.pem
#
# The esl files are requests to add a db key and a KEK.
# PK.pem is the certificate stored in PK, and might be useful for
# relocking a machine that had to have PK unset for some reason.
# PK.auth is a request to install PK.pem as the PK, and left for
# potential convenience.
#
# The esl files are unsigned, so once we're out of setup mode the
# files are harmless (can't be used to re-install old keys after
# replacement). They're just kept around so we can tell what keys
# the system was provisioned with later.
#
# If run with -d the additional files will be present
# dev_db_entry.esl
# PK.key  <--- DANGEROUS if present in a non-developer install
#
# dev_db_entry.esl is a request to add the PK to the db
# PK.key is the private key to allow signing custom code that the
# dev_db_entry will verify, or to remove the PK and disable
# secure boot.
#
# In developer mode, PK.key and PK.pem are merely copies of the
# provided developer keys from /usr/share/ renamed to feed the
# other parts of the process. They're kept should we ever decide
# to change them - we'd want the system to continue to have a
# copy of the ones that were installed in EFI.

OUTDIR=/var/lib/eos-payg-keys
KEYDIR=/usr/share/eos-payg
EFIDIR=/sys/firmware/efi/efivars

# Files that could result in security issues if they're recovered
# are created in /tmp on the assumption it will be tmpfs
TMPDIR=""

BAIL()
{
  echo "$0: $1" 1>&2

  if [ -d "$TMPDIR" ]; then
    rm -f "${TMPDIR}"/PKpriv "${TMPDIR}"/PK.key "${TMPDIR}"/securitylevel
    rmdir "${TMPDIR}"
  fi

  echo "Detritus may remain in $OUTDIR - please delete before trying again" 1>&2
  exit 1
}


ident=""
force=0
devmode=0
enroll_keys=0
print_help()
{
  echo Usage eos-payg-keygen.sh -i '<id>' "[OPTIONS]"
  echo Create secure boot keys for PAYG enrollment in $OUTDIR
  echo
  echo Arguments:
  echo -f'      ' Force - destroy and recreate output dir even if it exists
  echo -d'      ' Developer mode - keep a copy of the unlock key
  echo -h'      ' Help - print this text
  echo -i' <id> ' Store a unique id in the certificate - this is mandatory
  echo -e'      ' enroll keys into EFI firmware
  echo -s' <L>  ' Force security level to \'L\'
}

while getopts "fdhei:s:" option; do
  case "$option" in
  f) force=1
     ;;
  d) devmode=1
     ;;
  h) print_help
     exit 0
     ;;
  i) ident=$OPTARG
     ;;
  e) enroll_keys=1
     ;;
  s) override_security_level=$OPTARG
     ;;
  *) print_help
     exit 0
     ;;
  esac
done

# We need a unique identifier in the certificate to allow
# "emergency unlock" later
[ "$ident" != "" ] || BAIL "Unique id must be provided with the -i option"


if [ -d "$OUTDIR" ]; then
  [ $force -eq 1 ] || BAIL "Output directory $OUTDIR exists and may contain vital keys. run with -f to ignore"
  rm -rf "$OUTDIR"
fi

TMPDIR=$(mktemp -d)

mkdir -p "$OUTDIR"
[ $? -eq 0 ] || BAIL "Failed to create output directory $OUTDIR"

cd "$OUTDIR" || BAIL "Unable to switch to output directory $OUTDIR"

if [ $devmode -eq 0 ]; then
  openssl req -new -x509 -newkey rsa:2048 -subj "/CN=Endless PAYG PK/O=${ident}/" -keyout "${TMPDIR}"/PK.key -out PK.pem -days 7300 -nodes -sha256
  [ $? -eq 0 ] || BAIL "Failed to generate PK key and certificate"
else
  cp "${KEYDIR}"/EOS-PAYG-DEVELOPER-KEY-2019.pem PK.pem || BAIL "Failed to copy developer certificate"
  cp "${KEYDIR}"/EOS-PAYG-DEVELOPER-KEY-2019.key "${TMPDIR}"/PK.key || BAIL "Failed to copy developer key"
  # We keep a copy of the key for developers only.
  cp "${KEYDIR}"/EOS-PAYG-DEVELOPER-KEY-2019.key PK.key || BAIL "Failed to make persistent copy of developer key"
fi

cert-to-efi-sig-list PK.pem PK.esl
[ $? -eq 0 ] || BAIL "Failed to create PK signature list"

# Totally arbitrary timestamp. Normally the tool uses "current time",
# but we can't trust that the system clock is accurately set, and
# at least one efi fw implementation requires that the timestamp
# of a PK removal request be newer than the PK itself, so we specify
# times so we can satisfy that.
sign_time="2019-10-18 16:17:00"
now_seconds=$(date +%s)
sign_seconds=$(date -d "${sign_time}" +%s)
if [ "$sign_seconds" -ge "$now_seconds" ]; then
  # See https://phabricator.endlessm.com/T28259
  echo WARNING: The clock appears to be incorrect, PAYG may not function correctly!
fi

sign-efi-sig-list -t "${sign_time}" -k "${TMPDIR}"/PK.key -c PK.pem PK PK.esl PK.auth
[ $? -eq 0 ] || BAIL "Failed to sign PK signature list"

rm PK.esl

# We add efi attributes to the front of PK.key so we can store it
# in an efi variable.
printf "\x07\x00\x00\x00" > "${TMPDIR}"/PKpriv
[ $? -eq 0 ] || BAIL "Failed to create output file for PKpriv"
cat "${TMPDIR}"/PK.key >> "${TMPDIR}"/PKpriv
[ $? -eq 0 ] || BAIL "Failed to append to PKpriv"

rm -f "${TMPDIR}/PK.key"

# In developer mode we install the PK as a db key so the dev can sign things
if [ $devmode -eq 1 ]; then
  cert-to-efi-sig-list PK.pem dev_db_entry.esl
fi

# We could sign the KEK with the PK, but we don't need to in setup mode
# just make a signature list for efi-updatevar consumption
openssl x509 -in "${KEYDIR}"/EOS-PAYG-KEK-CA-2019.crt -inform DER -out KEK.pem -outform PEM
cert-to-efi-sig-list KEK.pem KEK.esl
[ $? -eq 0 ] || BAIL "Failed to create KEK siglist"
rm KEK.pem

# We can't sign with the KEK, but we don't need to - since we're in setup mode
# we just need the db entry as a siglist.
openssl x509 -in "${KEYDIR}"/EOS-PAYG-UEFI-CA-2019.crt -inform DER -out db_entry.pem -outform PEM
cert-to-efi-sig-list db_entry.pem db_entry.esl
[ $? -eq 0 ] || BAIL "Failed to create db siglist"
rm db_entry.pem

if [ $enroll_keys -eq 0 ]; then
  echo "$0": successfully completed without key enrollment
  exit 0
fi

# Note, we're not appending.
# we want to remove all other KEK and db entries
chattr -fi "${EFIDIR}"/KEK-8be4df61-93ca-11d2-aa0d-00e098032b8c
efi-updatevar -e -f "${OUTDIR}"/KEK.esl KEK || BAIL "Failed to set KEK"
chattr -fi "${EFIDIR}"/db-d719b2cb-3d3a-4596-a3bc-dad00e67656f
efi-updatevar -e -f "${OUTDIR}"/db_entry.esl db || BAIL "Failed to set db"

# Developers append their PK as an extra db key so they can self
# sign their work for testing.
if [ -f "${OUTDIR}"/dev_db_entry.esl ]; then
  chattr -fi "${EFIDIR}"/db-d719b2cb-3d3a-4596-a3bc-dad00e67656f
  efi-updatevar -a -e -f "${OUTDIR}"/dev_db_entry.esl db || BAIL "Failed to add developer key"
fi

chattr -fi "${EFIDIR}"/EOSPAYG_payload-d89c3871-ae0c-4fc5-a409-dc717aee61e7
mv -f "${TMPDIR}"/PKpriv "${EFIDIR}"/EOSPAYG_payload-d89c3871-ae0c-4fc5-a409-dc717aee61e7 2> /dev/null
# mv will throw two harmless errors trying to update permissions and time

# if PKpriv didn't disappear something is wrong.
if [ -f "${TMPDIR}"/PKpriv ]; then
  BAIL "Failed to prepare system for future unlock"
fi

# Once the PK is installed, we're out of setup mode and
# things will be locked down next boot (provided the bios
# is appropriate configured to secure boot).
#
# Don't need chattr here, because if there was a PK the
# previous steps would fail.
efi-updatevar -f "${OUTDIR}"/PK.auth PK || BAIL "Failed to set PK"

# Set the securitylevel variable to the installed eos-paygd's
# compile time security level.
chattr -fi "${EFIDIR}"/EOSPAYG_securitylevel-d89c3871-ae0c-4fc5-a409-dc717aee61e7
printf "\x07\x00\x00\x00" > "${TMPDIR}"/securitylevel
if [ -z ${override_security_level+x} ]; then
  /usr/lib/aarch64-linux-gnu/eos-paygd1 -s >> "${TMPDIR}"/securitylevel
else
  printf "\x$override_security_level" >> "${TMPDIR}"/securitylevel
fi
mv "${TMPDIR}"/securitylevel "${EFIDIR}"/EOSPAYG_securitylevel-d89c3871-ae0c-4fc5-a409-dc717aee61e7 2>/dev/null
if [ -f "${TMPDIR}"/securitylevel ]; then
  BAIL "Unable to set security level variable"
fi

# This should be empty now - we deleted one file and moved the other two.
rmdir "${TMPDIR}" || BAIL "Temporary directory not empty"

# Ok, we got this far, throw the switch to enable firmware protection.
# This should always be the very last step as it enabled full brick mode
# if something earlier went awry.
printf "\x07\x00\x00\x00\x00" > "${EFIDIR}"/EOSPAYG_active-d89c3871-ae0c-4fc5-a409-dc717aee61e7
# Make sure it's there - we can't just test for success from the printf because it will
# fail if EOSPAYG_active is already present and immutable, but it seems silly to chattr
# and rm just in case.  Let's just make sure the file is present.
if [ ! -f "${EFIDIR}"/EOSPAYG_active-d89c3871-ae0c-4fc5-a409-dc717aee61e7 ]; then
  BAIL "Failed to enable PAYG"
fi

echo "$0": successfully completed with key enrollment
